#!/usr/local/bin/perl
#
# OBSSUMM
#
# XMM ODF HTML summary generation script
#
# Author: Brendan Perry (bhp@star.le.ac.uk)
#
# Description: This script generates an HTML summary page from
#              the ODF summary ASCII file produced for each ODF.
#
# Usage: 
#       obssumm infile="ODF file" outfile="HTML file" \
#       logofile="logo.gif" (optional) -noclobber (optional)
#       usedataquality=boolean (optional) [true/yes false/no
#       determines whether or not to print the often long
#       data quality records]
#

sub obssumm {

    use strict;
    use English;
    use DAL;
    use SAS;
    use SAS::error;

    my ($taskname) = 'OBSSUMM';
    my ($date) = &date_time;
    my ($VERSION) = "4.2.6";
    my ($create) = "SAS task $taskname V $VERSION (Generated $date)";
    my ($msg,$code,$htmlfile,$obsid);
    my ($header_ok) = 0;
#
# use peDAL calls to get command line parameters
#
# new for 4.0...no longer uses infile as parameter (uses
# odfexpand for now, but should use perl OAL bindings when
# available
#
# 4.2...infile parameter restored at request of SSC-LUX 
# (better for the pipeline). Since it's only a PPS task anyway..
#
    my ($infile,$outfile,$usedq,$logo) = &parse_cl();
#
# check the ODF file for existence and readability
#
    my ($odf_ok) = &check_ODF($infile);
    if ($odf_ok) {
      SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Checked ODF file: $infile file is ok.");
    }
#
# open the ODF file
#
    my ($odffile) = &open_file($infile,"READONLY");
#
# read the ODF and fill odf reference hashes
#
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Reading ODF file $infile.");
    my ($obs_ref,$nfiles,$files_ref,$inst_ref,$expo_ref,$go_ref,$propobs_ref,$proppri_ref,$dataQ_ref,$dqok) = &read_ODF($odffile);
#
# close the opened and read ODF file:
#
  SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Closing ODF file $infile.");
    $odffile->close; 
#
# check the GIF logo file for existence and readability
#
    my ($logo_ok) = &check_logo($logo);
    if ($logo_ok) {
      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Checked GIF logo: $logo file is ok.");
    } else {
	$msg = "No LOGO banner written to HTML.";
	$code = "NoLogoBanner";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn
    }
#
# If at least have obs_ref, can write a (bad looking) html file:
#
    my $sub_title="<table border=\"0\" width=\"100\%\">\n";
    $sub_title .= "\t<tr align=\"left\" >\n";
    $sub_title .= "\t\t<th>Revolution</th>\n";
    $sub_title .= "\t\t<th>Target</th>\n";
    $sub_title .= "\t\t<th>Obs Length</th>\n";
    $sub_title .= "\t\t<th>Observer</th>\n";
    $sub_title .= "\t</tr>\n";
    $sub_title .= "\t<tr align=\"left\" >\n";
    $sub_title .= "\t\t<td>$obs_ref->{orbit}</td>\n";
    $sub_title .= "\t\t<td>$propobs_ref->{tg}</td>\n";
    $sub_title .= "\t\t<td>$propobs_ref->{len}</td>\n";
    $sub_title .= "\t\t<td>$go_ref->[0]</td>\n";
    $sub_title .= "\t</tr>\n";
    $sub_title .= "</table>\n";
#
# First: open HTML file and write header
#
    if (exists($obs_ref->{obsid})) {
      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing HTML Header.");
	($htmlfile) = &open_file($outfile,"WRITEONLY");
	$obsid = $obs_ref->{obsid};
	chomp ($obsid);
	my ($title) = "$obsid Observation Data File Summary";
	($header_ok) = &write_HTML_header($infile,$htmlfile,$title,$sub_title,$logo,$logo_ok,$obs_ref->{obsid});
	&write_generic($htmlfile,"<hr size=\"2\" noshade=\"noshade\" />\n");
    } else {
	$msg = "No HTML written, could not parse enough from infile.";
	$code = "NoHTMLwritten";
	chomp ($msg,$code);
      SAS::error($code,$msg);    # warn
    }
#
# write the GO information:
#
  SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Writing GO records.");
    if (@$go_ref) { 
	&write_HTML_go($htmlfile,@$go_ref);
	&write_generic($htmlfile,"<hr size=\"2\" noshade=\"noshade\" />\n");
    } else {
	$msg = "No Guest Observer Information table written to HTML.";
	$code = "NoGOInfoTbl";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn
    }
#
# write the proposal target information
#
  SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Writing Proposal Observation records.");
    if ($propobs_ref) { 
	&write_HTML_propobs($htmlfile,$propobs_ref);
	&write_generic($htmlfile,"<hr size=\"2\" noshade=\"noshade\" />\n");
    } else {
	$msg = "No Proposal Target Information table written to HTML.";
	$code = "NoPropTargetTbl";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn
    }	
#
# write the OBSERVATION section information:
#
    if ($obs_ref) { 
      SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Writing OBS records to table.");
	&write_HTML_obs($htmlfile,$obs_ref);
	&write_generic($htmlfile,"<hr size=\"2\" noshade=\"noshade\" />\n");
    } else {
	$msg = "No Observation Record table written to HTML.";
	$code = "NoObsRecTbl";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn
    }
#
# write the proposal priority information
#
  SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Writing Proposal Priority records.");
    if ($proppri_ref) { 
	&write_HTML_proppri($htmlfile,$proppri_ref,$inst_ref,$dataQ_ref);
	&write_generic($htmlfile,"<hr size=\"2\" noshade=\"noshade\" />\n");
    } else {
	$msg = "No Instrument Information table written to HTML.";
	$code = "NoInstInfoTbl";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn
    }
#
# write the Configuration information
#
  SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Writing Configuration records.");
    if ($dqok) {
	&write_HTML_config($htmlfile,$inst_ref,$expo_ref);
	&write_generic($htmlfile,"<hr size=\"2\" noshade=\"noshade\" />\n");
    } else {
	$msg = "No configuration table written to HTML.";
	$code = "NoConfigTbl";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn
    }
#
# write the Data Quality information
#
    if ($usedq) {
      SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Writing Data Quality records.");
	if ($dqok) {
	    &write_HTML_dataQ($htmlfile,$inst_ref,$dataQ_ref);
	    &write_generic($htmlfile,"<hr size=\"2\" noshade=\"noshade\" />\n");
	} else {
	    $msg = "No Data Quality table written to HTML.";
	    $code = "NoDataQTbl";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);    # warn
	}
    } else {
      SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Usedataquality param set to FALSE, no Data Quality Records written to HTML.");
    }
#
# create and write footer to HTML page
#
    &write_HTML_footer($htmlfile,$create);
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Wrote HTML footer.");
#
# close the open and written HTML file:
#
    $htmlfile->close;
  SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Closed HTML filehandle.");
}
#
# end sub obssumm
#
#=========================================================================
#=========================================================================
#
# subroutine section
#
#=========================================================================
# --
# PARSE_CL
# --

sub parse_cl {     # subroutine acts as command line parameter parser.

#
# TASKNAME infile="input filename" outfile="output HTML filename" \
#  logofile="logo.gif"
#
# infile and outfile are mandatory
#
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Parsing command line.");
    
    my ($useinfile) = booleanParameter("useinfile");
    my ($in) = '';
    if ($useinfile) {
	$in = stringParameter("infile");
      SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Getting infile from parameter.");
    } else {
      SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Getting infile from odfexpand.");
        $in = `odfexpand --odfrequest='odf(xmm, sum)'`;
    }
    my ($out) = stringParameter("outfile");
    my ($logo) = stringParameter("logofile");
    my ($usedq) = booleanParameter("usedataquality");
    chomp ($in,$out,$usedq,$logo);
    return ($in,$out,$usedq,$logo);
}
#
# end PARSE_CL
#
# --
# CHECK_ODF
# --
#
sub check_ODF {    # check the odf ascii summary file as input.

    my ($odf) = @_;
    my ($msg,$code,$ok);
#
# check to see if odf input file exists, is readable and
# is (ok for now) a Text file (stops binary file inputs but
# doesn't really work on FITS files because first 2048 bytes
# are usually ASCII, so file test operator does not catch it
# as binary (but some systems it DOES because it looks at a
# PERCENTAGE of 'binary chars'.
#
    $ok = 1;
    
    if (!(-e $odf)) {
	$msg = "File: $odf does not exist";
	$code = "ODFFileNotExist";
	chomp ($msg,$code);
	SAS::error($code,$msg);			
	$ok = 0;
    }
    
    if (!(-r $odf)) {
	$msg = "File: $odf not readable";
	$code = "NoReadODF";
	chomp ($msg,$code);
	SAS::error($code,$msg);			
	$ok = 0;
    }
    
    if (!(-T $odf)) {
	$msg = "File: $odf not a Text file";
	$code = "ODFNotText";
	chomp ($msg,$code);
	SAS::error($code,$msg);			
	$ok = 0;
    }
    
    return ($ok);
}
#
# end check_odf
#
# --
# OPEN_FILE
# --
sub open_file {   # open an HTML output file

    require IO::Handle;
    require FileHandle;

    my ($file,$mode) = @_;
    my ($msg) = "Unable to open file: $file.";
    my ($code);
    my ($clob) = DAL::clobber();    
    my ($fh) = new FileHandle;
    
    if ($fh) {
	if ($mode =~ /READ/) {              # open for read ONLY
	    $file = "<$file";  
	    if (!($fh->open("$file"))) {
		$code = "RdOnlyOpenFailure";
		chomp ($msg,$code);
	      SAS::error($code,$msg);    # exit with error
	    } else {
	      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Opened file $file READONLY");
	    }
	} elsif ($mode =~ /WRITE/) {        # open for write ONLY
	    if (-e $file) {
	      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Output file already exists.");
		if ( $clob ) {    # noclobber set to TRUE
		    $msg = "File exists but noclobber NOT set: Overwriting file.";
		    $code = "ClobberingFile";
		    chomp ($msg,$code);
		  SAS::warning($code,$msg);    # warn
		} 
		else {
		    # noclobber set to TRUE
		    $msg = "Cannot open WRITE with noclobber set.";
		    $code = "NoclobberOpenFailure";
		    chomp ($msg,$code);
		  SAS::error($code,$msg);    # exit with error
		}
	    }
	    $file = ">$file";  # overwrite
	    if (!($fh->open("$file"))) {
		$code = "WrtOnlyOpenFailure";
		chomp ($msg,$code);
	      SAS::error($code,$msg);    # exit with error
	    } else {
	      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Opened file $file WRITEONLY");
	    }
	}
	return ($fh);
    } else {
	$msg = "Unable to get new FileHandle";
	$code = "FileHandleFailure";
	chomp ($msg,$code);
      SAS::error($code,$msg);    # exit with error
    }
}
#
# end OPEN_FILE
#
# --
# CHECK_LOGO
# --
#
sub check_logo {   # check input logo file 

    my ($logo) = @_;
    my ($msg,$code,$ok);
#
# check the input logo file for existence, readability and
# if it's a Binary file. Not very vigorous test, but ok for now
# (if you're DUMB enough to put a non-gif/tiff/png file as
# a logo banner you get what you get!).
#
    if ($logo) {
	if (!(-e $logo)) {
	    $msg = "File: $logo does not exist.";
	    $code = "LogoFileNotExist";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);			
	    $ok = 0;
	} elsif (!(-r $logo)) {
	    $msg = "File: $logo not readable.";
	    $code = "NoReadLogoFile";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);			
	    $ok = 0;
	} elsif (!(-B $logo)) {
	    $msg = "File: $logo possibly not binary image file.";
	    $code = "LogoFileNotImage";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);			
	    $ok = 0;
	} else {
	    $ok = 1;    # file should be ok
	}
    } else {
	$ok = 0;
    }
    return ($ok);
}
#
# end check_logo
#
# --
# WRITE_HTML_HEADER
# --
sub write_HTML_header {       # start HTML file with HEAD,TABLE,BODY,etc

#
# subroutine to write the header information to the output www page.
# this includes the TITLE, colors, top anchor, etc
#
    my ($infile,$fh,$title,$sub_title,$logo,$logo_ok,$obsid) = @_;
    chomp ($title,$logo,$logo_ok,$obsid);
    SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Obs ID is: $obsid");
    my ($top) = "top";    # set anchor at top, links with bottom
    my ($bg) = "bgcolor=\"\#FFFFFF\"";   # white
    my ($text) = "test=\"\#0000EE\"";    # black
    my ($link) = "link=\"\#0000EE\"";    # black
    my ($vlink) = "vlink=\"\#551A8B\"";  # blueish?
    my ($alink) = "alink=\"\#FF0000\"";  # red?

    my $link_bar = &link_names;
    &write_generic($fh,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"); 
    &write_generic($fh,"<!DOCTYPE html\n");
    &write_generic($fh,"PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n");
    &write_generic($fh,"\"http://www.w3.org/TR/xhtml1/DTD/xhtml-transitional.dtd\">\n");

    &write_generic($fh,"<html>\n");
    &write_generic($fh,"<head><title>$title</title></head>\n");
    &write_generic($fh,"<body $bg $link $vlink $alink>\n");
    if ($logo_ok) {
	my ($imgsrc) = "<img src=\"$logo\" alt=\"XMM Science Survey Centre\" />";
	&write_generic($fh,"$imgsrc \n");
    }

    &write_generic($fh,"<hr noshade=\"noshade\" />\n");
    &write_generic($fh,"$link_bar\n");
    &write_generic($fh,"<hr noshade=\"noshade\" />\n");

    my ($temp) = &anchor($top);
    &write_generic($fh,"$temp<h1>$title</h1>\n");
    &write_generic($fh,$sub_title);
    &write_generic($fh,"<br />\n");
#    &write_generic($fh,"<hr noshade=\"noshade\" />\n");

}
#
# end WRITE_HTML_HEADER
#
# --
# WRITE_HTML_FOOTER
# --
sub write_HTML_footer {  # write the HTML footer info to the page bottom

    my ($fh,$create) = @_;
    my ($top) = "#top";
#    &write_generic($fh,"<br />\n");
    my ($line) = href($top,'Top of Page');  # reference to top anchor
    &write_generic($fh,"$line<br />\n");
    &write_generic($fh,"<h5>$create</h5>\n");
    &write_generic($fh,"</body>\n");
    &write_generic($fh,"</html>\n");
}
#
# end WRITE_HTML_FOOTER
#
sub anchor {           # create an anchor from input
    my ($name) = @_;
    my ($anchor) = "<a name=\"$name\"></a>";
    chomp $anchor;
    return ($anchor);
}

sub href {             # create an href from input

# <A HREF="obssumm.html">Observation Summary Page</A>

    my ($ref,$name) = @_;
    my ($href) = "<a href=\"$ref\">$name</a>";
    chomp $href;
    return ($href);
}

sub begin_table {      # start a new table

# <H2>Table Name:</H2> 

    my ($fh,$name,$border,$align,$width,$color,$colspan) = @_;
    chomp ($name);
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Begin Table: $name");
    my ($line) = "<table border=$border align=$align bgcolor=$color>\n";
# change to <pre> for version 5...

#    my ($line) = "<pre>\n";
    &write_generic($fh,$line);
    $line = "<th align=center colspan=$colspan>$name</th>";
    &write_generic($fh,$line);
}

sub end_table {        # complete a table

    my ($fh,$name) = @_;
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"End Table: $name");
    my ($line) = "</table>\n";
# change to </pre> for version 5...
#    my ($line) = "</pre>\n";
    &write_generic($fh,$line);
}

sub table_row {        # write a single table row

# <TD ALIGN=CENTER>$value</TD>

    my ($fh,$value,$type,$align) = @_;   # value, typeface B align CENTER
    chomp ($value,$type,$align);
# using pre, so just dump line and <br /> it.
    my ($tr) = "<td align=\"$align\"><$type>$value</$type></td>\n";
#    my ($tr) = "<td align=\"$align\"><$type>$value</$type></td>\n";
    &write_generic($fh,$tr);
}

sub date_time {          # create the file creation datestring

    my $time = localtime(time);
    my @time = split /\ /, $time;
    return "$time[2] $time[1] $time[4]";
}

#
# end write_line
#
#--
# WRITE_GENERIC
#--
sub write_generic {        # write a created HTML line

    my ($fh,$line) = @_;
    my ($msg) = "Failure to printf to output HTML file.";
    my ($code) = "WriteHTMLFailure";
    chomp ($msg,$code);
#
# exit with error if problem with printf
#
    if (!( printf ($fh $line) ) ) {
       	SAS::error($code,$msg);
    }
}
# --
# READ_ODF_FILE
# --
sub read_ODF {

    my ($odf) = @_;
    my ($prop,$cl,$i,$j,$msg,$code,$rest,$n,@arr);
    my (@flags,@obsinfo,@filelist,@instrument,@proposal,@dataQ);
    my ($obs_ref,$inst_ref,$propgo_ref,$propobs_ref,$proppri_ref);
    my ($file_ref,$dataQ_ref);
#
# slurp the file into array @arr
#
    my (@arr) = <$odf>;
    my ($n) = scalar (@arr);  # initial number of lines
    my ($nf,$texp) = 0;       # initialize num of files/exposures
#
# OBSERVATION or SLEW *should* be first string.
#
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Parsing OBSERVATION records.");
    for ($i = 0 ; $i < $n ; $i++) {
	chomp $arr[$i];
#
# bail from loop if found another section 'early' (can still recover
# SOME info from 'damaged infiles'.
#
	last if (($arr[$i] =~ /^FILES/) or ($arr[$i] =~ /^INSTRUMENT/) or ($arr[$i] =~ /^PROPOSAL/) or ($arr[$i] =~ /^DATA QUALITY/));
	if (($arr[$i] =~ /^OBSERVATION/) or ($arr[$i] =~ /^SLEW/)) {
	    @obsinfo = splice (@arr,$i,5);  # ICD says 5 records in OBS
	    last;
	}
    }   
#
# parse those lines of @obsinfo to make hash $obs_ref
#
    if (@obsinfo) {
	($obs_ref) = &parse_obs(@obsinfo);
    } else {
	$msg = "File possibly NOT ODF Summ: Unable to parse OBSERVATION keyword.";
	$code = "OBSKeysMissingInODF";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn, won't make obs table later
    }
#
# now get the FILES values
#
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Parsing FILES records.");
    for ($i = 0 ; $i < $n ; $i++) {
	chomp $arr[$i];
#
# bail from loop if found another section 'early' (can still recover
# SOME info from 'damaged infiles'.
#
	last if ($arr[$i] =~ (/^INSTRUMENT/) or $arr[$i] =~ (/^PROPOSAL/) or $arr[$i] =~ (/^DATA QUALITY/));
	if (($arr[$i] =~ /^FILES/)) {
	    ($nf,$rest) = split (/\s+\/\s+/,$arr[$i+1]);  # parse nfiles
	    if ( $nf =~ /\d+(?!\w+)/ ) {   # digits but not alphanumeric
		@filelist = splice (@arr,$i,$nf);
		last;
	    } else {
		$msg = "Failed to parse number of FILES records $nf";
		$code = "NfilesSetToZero";
		chomp ($msg,$code);
	      SAS::warning($code,$msg);
		$nf = 0;
	    }
	}
    }   
#
# parse those lines of @obsinfo to make hash $file_ref
#
    if (@filelist) {
	($files_ref) = &parse_file(@filelist);
    } else {
	$msg = "File possibly NOT ODF Summ: Unable to parse FILES keyword.";
	$code = "FILESKeysMissingInODF";
	chomp ($msg,$code);
	SAS::warning($code,$msg);
    }
#
# now get the INSTRUMENT values
#
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Parsing INSTRUMENT records.");

    for ($i = 0 ; $i < $n ; $i++) {
	chomp $arr[$i];
#
# bail from loop if found another section 'early' (can still recover
# SOME info from 'damaged infiles'.
#
	last if ($arr[$i] =~ (/^PROPOSAL/) or $arr[$i] =~ (/^DATA QUALITY/));
	if (($arr[$i] =~ /^INSTRUMENT/)) {
	    for ($j = $i ; $j < $n ; $j++ ) {
		chomp $arr[$j];
		last if ($arr[$j] =~ (/^PROPOSAL/) or $arr[$j] =~ (/^DATA QUALITY/));
		push (@instrument,$arr[$j]);
	    }
	}
    }
#
# parse those lines of @instrument to make hash $inst_ref
#
    if (@instrument) {
	($texp,$inst_ref,$expo_ref) = &parse_inst(@instrument);
    } else {
	$msg = "File possibly NOT ODF Summ: Unable to parse INSTRUMENT keyword.";
	$code = "INSTKeysMissingInODF";
	chomp ($msg,$code);
	SAS::warning($code,$msg);
    }
    if ($texp == 0) {  # calculate n exposures by hand
	@exps = grep /Exposure Id/,@arr;
	$texp = scalar(@exps);
    }
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Total exposures is $texp.");
#
# now get the PROPOSAL values
#
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Parsing PROPOSAL records.");
    for ($i = 0 ; $i < $n ; $i++) {
	chomp $arr[$i];
	last if ($arr[$i] =~ /^DATA QUALITY/);
	if ($arr[$i] =~ /^PROPOSAL/) {
	    for ($j = $i ; $j < $n ; $j++ ) {
		chomp $arr[$j];
		last if ($arr[$j] =~ /^DATA QUALITY/);
		push (@proposal,$arr[$j]);
	    }
	}
    }
#
# parse lines of @proposal,make hash $propgo_ref/$propobs_ref/$proppri_ref/
#
    if (@proposal) {
	($propgo_ref,$propobs_ref,$proppri_ref) = &parse_prop(@proposal);
    } else {
	$msg = "File possibly NOT ODF Summ: Unable to parse PROPOSAL keyword.";
	$code = "PROPKeysMissingInODF";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn and skip table later,
#	($propgo_ref,$propobs_ref,$proppri_ref) = 0; # by making hash 0
    }
#
# now get the DATA QUALITY values
#
    @dataQ = splice(@arr,$i);
    if ($dataQ[0] =~ /^DATA QUALITY/) {
	$n = scalar(@dataQ);
    }
#
# parse those lines of @dataQ to make hash $dataQ_ref
#
    my ($dqok) = 0;
    if (@dataQ) {
	($dataQ_ref,$dqok) = &parse_dataQ($texp,@dataQ);
    } else {
	$msg = "File possibly NOT ODF Summ: Unable to parse DATA QUALITY keywords.";
	$code = "DataQKeysMissingInODF";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn and skip table
    }
#
# Return to main with all summary file sections in hashes.
# These hashes become tables in HTML file...
#
# an HTML section will be created for each of these that contain
# correct data.
#
    return ($obs_ref,$nfiles,$files_ref,$inst_ref,$expo_ref,$propgo_ref,$propobs_ref,$proppri_ref,$dataQ_ref,$dqok);

}

sub parse_obs {   # parse the OBSERVATION subsection of infile
    
    my (@in) = @_;
    chomp @in;
    my (%obs,$line,$val,$rest,$msg,$code);

    my ($first) = shift (@in);
    chomp $first;

    if ($first =~ /^OBSERVATION/) {
    	if (@in) {
	    $line = shift (@in);
	    chomp $line;
	    ($val,$rest) = split (/\//,$line);
	    chomp $val;
	    $obs{obsid} = $val;   # observation ID
#
	    $line = shift (@in);
	    chomp $line;
	    ($val,$rest) = split (/\//,$line);
	    chomp $val;
	    $obs{orbit} = $val;   # orbit number
#
	    $line = shift (@in);
	    chomp $line;
	    ($val,$rest) = split (/\//,$line);
	    chomp $val;
	    $obs{start} = $val;   # start time 1998-02-18T13:30:00 format
#
	    $line = shift (@in);
	    chomp $line;
	    ($val,$rest) = split (/\//,$line);
	    chomp $val;
	    $obs{stop} = $val;    # stop time 1998-02-18T13:30:00 format
	}
    } elsif ($first =~ /^SLEW/) {
	$msg = "SLEW Observations not supported yet.";
	$code = "NoSLEWODFyet";
	chomp ($msg,$code);
	SAS::error($code,$msg);    # exit with error
    } else {
	$msg = "Cannot parse OBSERVATION section.";
	$code = "CannotParseOBS";
	chomp ($msg,$code);
	SAS::warning($code,$msg);  # warn, don't make table later.
    }

    return (\%obs);
}
#
# end parse_obj
#
#-- 
# PARSE_FILE
#--
sub parse_file {  # parse the FILES section of the infile.
#
# format, as per ODF ICD is A31,1X,A48
#
    my (@in) = @_;
    chomp @in;
    my (@files,$line,$val);
    
    my ($first) = shift (@in);
    chomp $first;
    
    if ($first =~ /^FILES/) {
	if (@in) {
	    $line = shift (@in); # shift off number of files
	    foreach $line (@in) {	    
		chomp $line;
		($val,$rest) = split (/\s+\/\s+/,$line);
		chomp $val;
#
# maybe throw out non-compliant filenames later...
#
		if ($val =~ /^\w{0,31}/) {
		    push (@files,$val);  # list of all files in ODF
		} else {
		  SAS::warning("BadFilenameFormat","File format for: $val not correct.");
		}
	    }
	}
    } else {
	$msg = "Unable to parse FILES section.";
	$code = "CannotParseFILES";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # exit with 
    }
    return (\@files);
}
#
# end parse_file
#
#--
# PARSE_INST
#--
sub parse_inst {   
#
# parse the INSTRUMENT/EXPOSURE/CONFIGURATION sections of infile
#    
  my (@arr) = @_;
  my (@file,$line,$val,$nf,$n,$i,$j,$k,$expkey,$exp,$full);
  my ($cl,$inst,$active,$nexp,$np,$par,$texp,$ninst);
  my (%config,@expvals);
  my (@exppars) = ("EXPNO","SCHED","TYPE","MISSEXPID","SCHED_START","SCHED_STOP","ACT_START","ACT_STOP");
  chomp @arr;
  $nlines = scalar(@arr);
  
 BIGLOOP:  for (;;) {
     $cl = shift @arr;
     last if (scalar(@arr) < 1);   # failsafe bailout
     if ($cl =~ /^INSTRUMENT/) {
	 $ninst++;              # increment number of instruments
	 last if ($ninst > 6);  # can't have more than 6 instruments!
	 $cl = shift @arr;
	 if ($cl =~ /^(M1|M2|R1|R2|OM|PN|m1|m2|r1|r2|om|pn)/) {
	     chomp $cl;    # instrument (M1)
	     ($inst,$full) = split (/\s+\/\s+/,$cl);
	 }
	 $cl = shift @arr;
	 ($val,$comment) = split (/\s+\/\s+/,$cl);
	 $config{$inst}{active} = $val;    # active flag
	 $cl = shift @arr;
	 ($nexp,$comment) = split (/\s+\/\s+/,$cl); 
	 if (!($nexp =~ /\d+/)) { $nexp = 0; }   # nexp must be digits
	 $config{$inst}{nexp} = $nexp;  # n exp's for this instrument
	 $texp += $nexp;
       SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Processing $inst instrument information");
       SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"$inst seems to have $nexp exposures");
	 
	 for (;;) {     # peel off comments until get to first EXPOSURE
	     #	      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"cl is $cl");
	     $cl = shift (@arr);
	     if ($cl =~ /^INSTRUMENT/) {      # failsafe if no exposure recs
		 unshift @arr, $cl;
		 next BIGLOOP;
	     }
	     last if ($cl =~ /^EXPOSURE/);
	     last if (scalar(@arr) < 1);   # failsafe bailout
	 }
#
# now get EXPOSURE records ($val)
#
	 if ($nexp) {   # if there are any exposures for this instrument...
	     
	     for ($exp = 0 ; $exp < $nexp ; $exp++ ) {
		 #		    next if ($cl =~ /^EXPOSURE/);
		 #	  $cl = shift (@arr);
		 #
		 # shift off all the exposure information
		 #
		 for ($ne = 0 ; $ne < 9 ; $ne++) {
		     $cl = shift (@arr);
		     chomp $cl;
		     ($expvals[$ne],$comment) = split (/\s+\/\s+/,$cl);
		     last if ($cl =~ /^CONFIG/);      # failsafe
		     if ($cl =~ /^INSTRUMENT/) {      # failsafe also
			 unshift @arr, $cl;
			 next BIGLOOP;
		     }
		 }
		 $expid = $expvals[3];
#	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"inst = $inst expid = $expid");
		 splice (@expvals,4,1);
		 $expo{$inst}{$expid} = $expid;
		 $k = 0;
		 #
		 # exppars: "EXPNO","SCHED","TYPE","MISSEXPID","SCHED_START",
		 #          "SCHED_STOP","ACT_START","ACT_STOP"
		 #
		 foreach $expkey (@exppars) {
		     chomp $expkey;
		     $config{$inst}{$expid}{$expkey} = $expvals[$k];
		     $k++;
		 } 
		 for (;;) {   # peel off comments until get to first CONFIG
		     $cl = shift (@arr);
		     last if ($cl =~ /^CONFIGURATION/);
		     if ($cl =~ /^INSTRUMENT/) {      # failsafe also
			 unshift @arr, $cl;
			 next BIGLOOP;
		     }
		     last if (scalar(@arr) < 1);   # failsafe bailout
		 }
		 #
		 # no config record for this exposure if next thing seen is EXPOSURE
		 #
		 # loop through the configuration records config_rec
		 #
		 if ($cl =~ /^CONFIGURATION/) {
		     $cl = shift (@arr);	# shift off number of config params, $ncon
		     ($ncon,$comment) = split (/\s+\/\s+/,$cl); 
		     if ($ncon > 0) {    # if there are config params...
			 for ($nc = 0 ; $nc < $ncon ; $nc++) {
			     $cl = shift (@arr);
			     if ($cl =~ /^INSTRUMENT/) {      # failsafe also
				 unshift @arr, $cl;
				 next BIGLOOP;
			     }
			     chomp $cl;         # split on "/" first
			     ($line,$comment) = split (/\s+\/+\s+/,$cl);
			     #		SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"line is $line");
			     chomp $line;       # split on "="
			     ($par,$val) = split (/\s+\=\s+/,$line);
			     chomp $par;
			     $config{$inst}{$expid}{$par}=$val;
			     
#		if ($par =~ /^MODE/) {
#		  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"inst = $inst expid = $expid par = $par $config{$inst}{$expid}{$par} val is $val");
#		}
			 }			   
		     }
		 }
		 for (;;) {    # peel off comments until get to new EXPOSURE
		     $cl = shift (@arr);
		     if ($cl =~ /^INSTRUMENT/) {
			 unshift @arr, $cl;
			 last;
		     }
		     last if ($cl =~ /^EXPOSURE/);
		     last if (scalar(@arr) < 1);   # failsafe bailout
		 }
		 
	     }
	 }
     } else {
	 if (!($i)) {
	     $msg = "Unable to parse INSTRUMENT section.";
	     $code = "CannotParseINSTRUMENT";
	     chomp ($msg,$code);
	   SAS::warning($code,$msg);    # exit with 
	     $i++;	    
	 }
     }
 }
  return ($texp,\%config,\%expo);    
}
#
# end parse_inst
#
#--
# PARSE_PROP
#--
sub parse_prop {   # parse the PROPOSAL section of the infile.
#
    my (@prop) = @_;
    my ($i) = 0;
    my ($val,$rest,$line,$coo);
    my (%obs,%pri,@go);
#
# first, parse the GO info...
#
    shift (@prop);

    for ($i=0;$i<9;$i++) {   # 10 lines per ODF ICD for GO info.
	$line = shift (@prop);
	chomp $line;
	push (@go,$line);
    }
    
    $line = shift (@prop);
    chomp $line;
    if ($line) { ($obs{ao},$rest) = split (/\s+\/\s+/,$line); } # ao number

    $line = shift (@prop);
    chomp $line;
    if ($line) { ($obs{st},$rest) = split (/\s+\/\s+/,$line); } # science type

    $line = shift (@prop);
    chomp $line;
    if ($line) { ($obs{tg},$rest) = split (/\s+\/\s+/,$line); } # target
    
    $line = shift (@prop);
    chomp $line;
    if ($line) {  # target RA
      ($coo,$rest) = split (/\s+\/\s+/,$line);
      chomp ($coo);
      $obs{ra} = &convert_coords($coo,"RA");
    }

    $line = shift (@prop);
    chomp $line;
    if ($line) {  # target DEC
      ($coo,$rest) = split (/\s+\/\s+/,$line);
      chomp ($coo);
      $obs{dec} = &convert_coords($coo,"DEC");
    }

    $line = shift (@prop);
    chomp $line;
    if ($line) { # obs duration
      ($coo,$rest) = split (/\s+\/\s+/,$line);
      $obs{len} = sprintf("%d seconds",$coo);
    } 

    $line = shift (@prop);
    chomp $line;
    if ($line) { # altnames
	($coo,$rest) = split (/\s+\/\s+/,$line);
 	if ($coo =~ /\S+/) {
	    $obs{alt} = $coo;
	} else { 
	    $obs{alt} = "N\/A";
	}
    } # altnames

    $line = shift (@prop);
    chomp $line;
    if ($line) { # boresight RA
	($coo,$rest) = split (/\s+\/\s+/,$line);
	$obs{bra} = &convert_coords($coo,"RA");
    }

    $line = shift (@prop);
    chomp $line;
    if ($line) {  # boresight DEC
	($coo,$rest) = split (/\s+\/\s+/,$line);
	$obs{bdec} = &convert_coords($coo,"DEC");
    }

    $line = shift (@prop);
    chomp $line;
    if ($line) { ($obs{lpa},$rest) = split (/\s+\/\s+/,$line); } # lowPA

    $line = shift (@prop);
    chomp $line;
    if ($line) { ($obs{upa},$rest) = split (/\s+\/\s+/,$line); } # upPA

    $line = shift (@prop);
    chomp $line;
    if ($line) { ($obs{scpa},$rest) = split (/\s+\/\s+/,$line); } # scPA

    $line = shift (@prop);
    chomp $line;

    if ($line) { ($pri{M1},$rest) = split (/\s+\/\s+/,$line);    } # prioritym1

    $line = shift (@prop);
    chomp $line;
    if ($line) { ($pri{M2},$rest) = split (/\s+\/\s+/,$line); } # prioritym2

    $line = shift (@prop);
    chomp $line;
    if ($line) { ($pri{PN},$rest) = split (/\s+\/\s+/,$line); } # prioritypn

    $line = shift (@prop);
    chomp $line;
    if ($line) { ($pri{R1},$rest) = split (/\s+\/\s+/,$line); } # priorityr1

    $line = shift (@prop);
    chomp $line;
    if ($line) { ($pri{R2},$rest) = split (/\s+\/\s+/,$line); } # priorityr2

    $line = shift (@prop);
    chomp $line;
    if ($line) { ($pri{OM},$rest) = split (/\s+\/\s+/,$line); } # priorityom

    return (\@go,\%obs,\%pri);

}
#
# end parse_prop
#
# --
# PARSE_DATAQ
# --
sub parse_dataQ { # parse the DATA QUALITY section of the infile

    my ($texp,@arr) = @_;
    my (%dq,$pack,$nevt,$npms,$ntel,$ttel,$nool,$sch);
    my ($i,$ntot,$line,$val,$rest,$cl,$inst,$comment,$expid) = 0;
    my ($ok) = 0;
    chomp @arr;
    $cl = shift (@arr);

    if ($cl =~ /^DATA QUALITY/) {
	$ok = 1;
      DQ: for ($j = 0 ; $j < $texp ; $j++) {
	  $val = 0;   # zero out whole array each time in case no vals
	  for ($i = 0 ; $i < 9 ; $i++ ) {
	      $cl = shift @arr;
	      ($val[$i],$comment) = split (/\s+\/\s+/,$cl);
	      chomp ($val[$i]);
	  }
	  shift (@arr);
#
# check the value was properly set. If lines missing from DATA QUALITY
# dump that exposure and go to the next one!
#
	  if (!($val[1] =~ /\w{12,15}/)) {
#	      if (!($k)) { 
#		  $msg = "Data Quality Records not in correct format.";
#		  $code = "BadDataQualityRecord";
#		  chomp ($msg,$code);
#		SAS::warning($code,$msg);
#		  $k++;
#	      }
	      while (scalar(@arr) > 0) {
		  $cl = shift (@arr);
		  next DQ if ($cl =~ /DATA QUALITY/);
	      }
	      $val[0..8] = 0;
	  }	
#
# fill hash here (will fill with zeros if invalid).
#	    
	  for ($i = 0; $i < @val; $i++) {
	      $val[$i] =~ s/^\s+//; # strip leading white
	      $val[$i] =~ s/\s+$//; # strip trailing white
	  }

	  $expid = $val[1];
	  if (!($val[0] =~ /(R1|R2|r1|r2|M1|M2|m1|m2|OM|om|SC|sc)/)) { $val[0] = '-'; }
	  if (!($val[2] =~ /(\d+)/)) { $val[2] = '-'; }
	  if (!($val[3] =~ /(\d+)/)) { $val[3] = '-'; }
	  if (!($val[4] =~ /(\d+)/)) { $val[4] = '-'; }
	  if (!($val[5] =~ /(\d+)/)) { $val[5] = '-'; }
	  if (!($val[6] =~ /(\d+)/)) { $val[6] = '-'; }
	  if (!($val[7] =~ /(\d+)/)) { $val[7] = '-'; }
	  if (!($val[8] =~ /(^\w{1,2})$/)) { $val[8] = '?'; }

	  $dq{$expid}{inst} = $val[0];  # instrument
	  $dq{$expid}{pack} = $val[2];  # failed packets
	  $dq{$expid}{nevt} = $val[3];  # n events
	  $dq{$expid}{npms} = $val[4];  # pms reports
	  $dq{$expid}{ntel} = $val[5];  # n telem drops
	  $dq{$expid}{ttel} = $val[6];  # time of telem drops
	  $dq{$expid}{nool} = $val[7];  # n ool
	  $dq{$expid}{sch}  = $val[8];  # sched/unsched obs
      }
    } else {
	$msg = "Unable to parse DATA QUALITY section.";
	$code = "CannotParseDATAQUALITY";
	chomp ($msg,$code);
      SAS::warning($code,$msg);
    }
    return (\%dq,$ok);
}
#
# end parse_dataQ
#
##### WRITE each subsection of file to HTML #####

#-- 
# WRITE_HTML_GO
#--
sub write_HTML_go {   # write the Guest Observer info to the HTML file.

    my ($fh,@go) = @_;
    my ($line,$name,$msg,$code);

    if (@go) {
	$name = "Guest Observer Information";
	chomp $name;
	&write_generic($fh,"<h3>$name</h3>\n");
	&write_generic($fh,"<pre>\n");

	foreach $line (@go) {
	    $line =~ s/#.*//;            # no comments
	    $line =~ s/^\s+//;           # no leading white
	    $line =~ s/\s+$//;           # no trailing white
	    $line =~ s/\s+/ /g;          # no middle white +
	    next unless length $line;
	    chomp $line;
	    next unless length $line;
	    &write_generic($fh,"<tt>$line</tt>\n");
	}
	&write_generic($fh,"</pre>\n");
    } else {
	$msg = "Unable to parse Guest Observer section.";
	$code = "CannotParseGOinfo";
	chomp ($msg,$code);
	SAS::warning($code,$msg);
    }
}
# 
# end write_html_go
#
#-- 
# WRITE_HTML_PROPOBS
#--
sub write_HTML_propobs {   # write the Proposal target info

    my ($fh,$prop) = @_;
    my ($key,$val,$name);
    my (@strings) = ("Target","RA","Dec","Obs Length","Alt Names","Boresight RA","Boresight Dec","Lower Pos Ang","Upper Pos Ang","SC Pos Ang","AO Number","Science Type");
    my(@keys) = ("tg","ra","dec","len","alt","bra","bdec","lpa","upa","scpa","ao","st");

    $name = "Proposal Target Information";
    chomp $name;
    &write_generic($fh,"<h3>$name</h3>\n");
    &write_generic($fh,"<pre>\n");
    
    for ($i = 0 ; $i < @keys ; $i++ ) {
	$name = $strings[$i];
	$val = $prop->{$keys[$i]};
	$val =~ s/#.*//;            # no comments
	$val =~ s/^\s+//;           # no leading white
	$val =~ s/\s+$//;           # no trailing white
	$name = sprintf("%-15s",$name);
	$val = sprintf("%-65s",$val);
	next unless length $val;
	&write_generic($fh,"<tt>$name: $val</tt>\n");
    }
    &write_generic($fh,"</pre>\n");
}
# 
# end write_html_propobs
#
#-- 
# WRITE_HTML_OBS
#--
sub write_HTML_obs {   # write the ODF OBSERVATION record to the HTML file.
    
    my ($fh,$obs) = @_;
    my ($line,$name,$i,$val);
    my (@strings) = ("Observation ID","Revolution","Scheduled Start","Scheduled Stop");
    my (@keys) = ("obsid","orbit","start","stop");
    $name = "Observation Record";
    chomp $name;
    &write_generic($fh,"<h3>$name</h3><br />\n");
    &write_generic($fh,"<pre>\n");

    for ($i=0;$i<@strings;$i++) {
	chomp ($strings[$i],$keys[$i]);
	$name = $strings[$i];
        $val = $obs->{$keys[$i]};
	$strings[$i] =~ s/#.*//;            # no comments
	$strings[$i] =~ s/^\s+//;           # no leading white
	$strings[$i] =~ s/\s+$//;           # no trailing white
	$val =~ s/#.*//;               # no comments
	$val =~ s/^\s+//;              # no leading white
	$val =~ s/\s+$//;              # no trailing white
	$strings[$i] = sprintf("%-15s",$strings[$i]);
	$keys[$i] = sprintf("%-65s",$val);
	next unless length $;
	&write_generic($fh,"<tt>$strings[$i]: $keys[$i]</tt>\n");
    }
    &write_generic($fh,"</pre>\n");
#    &end_table($fh,$name);    
}
#
# end write_html_obs
#
#-- 
# WRITE_HTML_PROPPRI
#--
sub write_HTML_proppri {   # write the Proposal priority info

    my ($fh,$prop,$iref,$dqref) = @_;
    my ($key,$val,$tab,$name,$line,$act,$exp);
    my (@keys) = ("M1" , "M2" , "OM" , "PN" , "R1" , "R2");
    my (@insts) = ("MOS 1" , "MOS 2" , "OM" , "PN" , "RGS 1" , "RGS 2");
    my (@cols) = ("Instrument" , "Active" , "Exposures" , "Priority");

    $tab = "Instrument Information";
    chomp $tab;
    &write_generic($fh,"<h3>$tab</h3>\n");

    $line = sprintf("%-10s  %-6s  %-9s  %-8s",$cols[0],$cols[1],$cols[2],$cols[3]);
    chomp $line;
    &write_generic($fh,"<h4><pre>$line</pre></h4>\n");
    &write_generic($fh,"<pre>\n");
    for ($i = 0 ; $i < @keys ; $i++ ) {  # keys are inst codes
#
# $name is the current instrument
#
	$name = $insts[$i];
#
# now get the value for active flag
#
	$act = $iref->{$keys[$i]}{active}; # key is instrument (e.g. R1)
	chomp ($act);
	$act =~ s/^\s+//; # leading white
	$act =~ s/\s+$//; # trailing white
	$act = '-' unless length($act);
	chomp $act;
#
# if no valid active flag, calculate it from data quality (dq) records
#
	if ($act eq '-' ) { 
	  SAS::warning("NoActiveFlag","No Active Flag found for $keys[$i] - Determining from Quality Records.");
	    $act = "N";
	   Findit: foreach $key (keys %{$dqref}) {
		if ($dqref->{$key}->{inst} =~ /$keys[$i]/) { 
		    $act = "Y";
		    last Findit;
		}
	    }
	}
	
	($m1exp,$m2exp,$r1exp,$r2exp,$omexp) = 0;
	$exp = $iref->{$keys[$i]}{nexp};  # number of exposures
	chomp ($exp);
	$exp =~ s/^\s+//; # leading white
	$exp =~ s/\s+$//; # trailing white
	$exp = "-" unless length($exp);

	if ($exp eq '-' ) { # calculate nexposures from dq records
	    $exp = 0;
	  SAS::warning("NoNexp","No Nexposures found for $keys[$i] - Determining from Quality Records.");
	   Findit: foreach $key (keys %{$dqref}) {
		if ($dqref->{$key}->{inst} eq $keys[$i]) { 
		    $exp++;
		}
	    }
	}

	$prior = $prop->{$keys[$i]};   # priority
	chomp ($prior);
	$prior =~ s/^\s+//; # leading white
	$prior =~ s/\s+$//; # trailing white
	$prior = '?' unless length($prior);
	if ($prior eq '?') {
	  SAS::warning("NoPriority","No Priority value found for $keys[$i] - Set to ? ");
	}
	#	&table_row($fh,$prior,"B","CENTER");
	$line = sprintf("%-10s   %-6s   %-9s   %-8s",$name,$act,$exp,$prior);
	next unless length $line;
	&write_generic($fh,"<tt>$line</tt>\n");
      }
    &write_generic($fh,"</pre>\n");
  }
# 
# end write_html_proppri
#
#-- 
# WRITE_HTML_CONFIG
#--
sub write_HTML_config {   # write the Config info

    my ($fh,$iref,$eref) = @_;
    my ($key,$tab,$name);
    my ($colline1,$colline2,$colline3);
    my (@keys) = ("SCHED","MODE","FILTER","FILTER_WHEEL","SCHED_START","SCHED_STOP");

    my @colnames = ("Inst","ExpID","Sched?","Mode","Filter","Position","Actual Start","Actual Stop");

    $tab = "Exposure and Configuration Information";
    chomp $tab;
    &write_generic($fh,"<h3>$tab</h3>\n");

    $nkeys = 6;
    $ninst = 0;
    for ($i=2;$i<8;$i++) {
	$key = $keys[$i-2];
	$collengths{$key} = length($colnames[$i]);
#      SAS::message($SAS::AppMsg,$SAS::SparseMsg,"col: $key width: $collengths{$key} col: $colnames[$i]");
    }
#
# run through hash, pad out/clean values, fill print array
#
    foreach $instr (sort keys %$iref) {   # sort on instrument name
	chomp $instr;         # key is inst here
	foreach $expid (sort values %{ $$eref{$instr} } ) { #sort on expID
	    chomp $expid;
	    $newexpid = substr($expid,12);
	    $all[$ninst][0] = $instr;
	    $all[$ninst][1] = $newexpid;  # shortened expID (e.g. S009)
	    for($i=0;$i<$nkeys;$i++) {
		$col = $keys[$i];
		chomp $col;
		$val = $iref->{$instr}{$expid}{$col};
		$val =~ s/^\s+//;      # leading white
		$val =~ s/\s+$//;      # trailing white
		$val =~ s/\.000+$//;   # trailing ".000"
		$val = "-" unless length($val);
		$foo = $i+2;
		$all[$ninst][$foo] = $val;
#
# width will be the wider of largest VALUE width for a
# given column and the width of the column TITLE. See above
#
		if (length($val) > $collengths{$col} ) {
		    $collengths{$col} = length($val);
		}
	    }
	    $ninst++;
	}	
    }
#
# now sort @all array (on instrument then start time)
#
    @sortedall = sort { $a->[0] cmp $b->[0] || $a->[6] cmp $b->[6] } @all;
#
# dump each row of @all array into HTML
#
#    foreach $key (sort keys %collengths) {
#      SAS::message($SAS::AppMsg,$SAS::SparseMsg,"NEW col: $key width: $collengths{$key}");
#    }

    $layout = " %-5s %-5s";
    foreach $key (@keys) {
	$layout .= " %-".$collengths{$key}."s ";
    }

  SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Table layout is: $layout");

    $colline2 = sprintf("$layout",@colnames);
    chomp ($colline1,$colline2);
    &write_generic($fh,"<h4><pre><font size=\"-1\">$colline1</font></pre></h4>\n");
    &write_generic($fh,"<h4><pre><font size=\"-1\">$colline2</font></pre></h4>\n");
    &write_generic($fh,"<pre>\n");

    for ($i=0;$i<$ninst;$i++) { # instruments
	for ($j=0;$j<8;$j++) {  # "full" column names
#
# if width of column element less than designated max width, pad it
# out (so columns all line up properly)
#
	    if ($j > 1) {
		$val_len = length($sortedall[$i][$j]);
		$col_len = $collengths{$keys[$j-2]};
		if ($val_len < $col_len) {		    
		    $lll = length($sortedall[$i][$j]);
		    $lyt = "%-".$col_len."s";
		    $padname = sprintf("$lyt",$sortedall[$i][$j]);
		} else {
		    $padname = $sortedall[$i][$j];
		}
	    } else {
		$padname = sprintf("%-5s",$sortedall[$i][$j]);
	    }
	    push (@newall,$padname);
	}
	$line = sprintf("$layout",@newall);
	&write_generic($fh,"<font size=\"-1\"><tt>$line</tt></font>\n");
	undef @newall;
    }
    &write_generic($fh,"</pre>\n");
}
# 
# end write_html_config
#
#-- 
# WRITE_HTML_DATAQ
#--
sub write_HTML_dataQ {   # write the Data Quality info

    my ($fh,$iref,$dqref) = @_;
    my ($key,$tab,$name);
    my ($colline1,$colline2,$colline3);
    my (@keys) = ("inst" , "pack" , "nevt" , "npms" , "ntel" , "ttel" , "nool" , "sch");

    $tab = "Data Quality Information";
    chomp $tab;
    &write_generic($fh,"<h3>$tab</h3>\n");

    $colline1 = sprintf ("%s","                                                                   Duration of");
    $colline2 = sprintf ("%s","                    Failed     Event       PMS       Telemetry      Telemetry     Telemetry     Scheduled/");
    $colline3 = sprintf ("%s","Exp ID    Inst     Packets    Reports     Errors       Drops          Drops         OOL's       Unscheduled");

    chomp ($colline1,$colline2,$colline3);
    &write_generic($fh,"<pre>\n");
    &write_generic($fh,"<font size=\"-1\">$colline1</font>\n");
    &write_generic($fh,"<font size=\"-1\">$colline2</font>\n");
    &write_generic($fh,"<font size=\"-1\">$colline3</font>\n");
    &write_generic($fh,"</pre>\n");
    &write_generic($fh,"<br />\n");

    &write_generic($fh,"<pre>\n");
    foreach $expid (sort keys %$dqref) {         # write exp id first
      SAS::message($SAS::AppMsg,$SAS::SparseMsg,"dataq expid $expid");
	chomp $expid;         # key is exp id
	$newexpid = substr($expid,12);
	chomp $newexpid;
	$instr = $$dqref{$expid}{inst};

	foreach $col ( @keys ) {  # each exp ID has key vals; write them
	    chomp $col;
	    $val = $$dqref{$expid}{$col};
	    $val =~ s/^\s+//; # leading white
	    $val =~ s/\s+$//; # trailing white
	    $val = "-" unless length($val);
	    push (@all,$val);

	}
	$line = sprintf("%4s      %3s    %7s   %7s    %7s        %7s        %7s     %7s        %7s",$newexpid,$all[0],$all[1],$all[2],$all[3],$all[4],$all[5],$all[6],$all[7]);
	&write_generic($fh,"<font size=\"-1\"><tt>$line</tt></font>\n");
	undef @all;
    }
    &write_generic($fh,"</pre>\n");
}
# 
# end write_html_dataq
#
sub convert_coords { 
    #
    # string comes in as hh.dddd (hours and decimal hours) for RA
    # string comes in as dd.dddd (degrees and decimal degrees) for Dec
    # make a string HH:MM:SS.ddd HH.ddd DDD.dd if RA
    # make a string +/-DD:MM:SS.ddd (+/-nn.nn) if Dec
    #
    my ($value,$type) = @_;
    my ($str) = '';
    my ($frac,$m,$mm,$ss,$newfrac) = 0;
    my ($HH,$MM,$SS,$ddd) = 0;  
    
    my ($int) = int ($value);            
    if ($type =~ /RA/) {   
	my ($HH) = int($value);          # integer hours
	$frac = abs($value) - abs($HH);  # remaining fractional hours.
	$mm = ($frac * 60.);               # decimal minutes.
	$MM = int($mm);                  # integer minutes.
	$frac = abs($mm) - abs($MM);     # remaining fractional minutes.
	$ss = ($frac * 60.00);               # decimal seconds. 
	$str = sprintf("%2.2dh%2.2dm%-5.2fs (%8.4f hours)",$HH,$MM,$ss,$value);
    } else {
	my ($dd) = $int;               # $dd = -12
	$frac = abs($value) - abs($dd);
	$m = $frac * 60.;        
	$mm = int($m);
	$newfrac = abs($m) - abs($mm); 
	$ss = $newfrac * 60.;        # $secs = int(24.12) = 24
	$str = sprintf("%2.2dD%2.2dm%-5.2fs (%8.4f deg)",$dd,$mm,$ss,$value);
    }
    $str =~ s/^\s+//;      # strip leading white
    $str =~ s/\s+$//;      # strip trailing white
    $str =~ s/\(\s+/\(/g;  # no leading white before digits in ()'s
    return ($str);
    
}

#
# Create Link Filenames
#
sub link_names {
    my ($fn) = @_;
    my @bits = split /\_/, $fn;
    $fn = "P".$bits[1];

    my $bordercol="#000099";
    my $highback="#000099";
    
    @arr = ('OBS Summary','PPS Summary','EPIC Summary','OM Summary','RGS Summary','Catalogue Summary');
    my @links = (
		  "$fn"."OBX000SUMMAR0000.HTM"
		 ,"$fn"."OBX000PPSSUM0000.HTM"
		 ,"$fn"."EPX000SUMMAR0000.HTM"
		 ,"$fn"."OMX000SUMMAR0000.HTM"
		 ,"$fn"."RGX000SUMMAR0000.HTM"
		 ,"$fn"."CAX000XCORRE0000.HTM"
		 );
    
    my $link_bar = "<table width=\"90\%\" align=\"center\" border=\"0\" cellspacing=\"0\" cellpadding=\"3\">";
    $link_bar .= "\t<tr><td colspan=\"6\">&nbsp;</td></tr>\n";
    $link_bar .= "\t<tr align=\"center\">\n";
    $link_bar .= "\t\t<td width=\"15\%\" align=\"center\">$arr[0]</td>\n";
    for(my $i=1;$i<6;$i++) {
	$link_bar .= "\t\t<td width=\"15\%\" align=\"center\"><a href=\"$links[$i]\">$arr[$i]</a></td>\n";
    }
    
    $link_bar .= "\t</tr>\n";
    $link_bar .= "\t<tr><td colspan=\"6\">&nbsp;</td></tr>\n";
    $link_bar .= "</table>\n";

    return $link_bar;
}

##### end sub obssumm #####
